from unittest.mock import create_autospec
from unittest.mock import patch

import pytest
from packaging.utils import canonicalize_name
from packaging.version import Version
from pypi_simple import DistributionPackage
from pypi_simple import ProjectPage

from piplexed.pipx_venvs import PackageInfo
from piplexed.pypi_info import PackageVersions
from piplexed.pypi_info import find_outdated_packages
from piplexed.pypi_info import get_latest_version
from piplexed.pypi_info import get_pypi_versions


@pytest.mark.parametrize(
    "packages_data, non_pre_release, expected",
    [
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0b0", "sdist", False),
            ],
            True,
            Version("1.0.0"),
            id="no pre-releases",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0b0", "sdist", False),
            ],
            False,
            Version("2.0b0"),
            id="pre-release allowed",
        ),
    ],
)
def test_get_latest_version_stable_flag(packages_data, non_pre_release, expected):
    packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in packages_data
    ]
    assert get_latest_version(packages, stable=non_pre_release) == expected


@pytest.mark.parametrize(
    "packages_data, non_pre_release, expected",
    [
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "sdist", False),
            ],
            True,
            Version("2.5.0"),
            id="standard sdists",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "sdist", True),
            ],
            True,
            Version("2.0.0"),
            id="yanked sdists",
        ),
        pytest.param(
            [
                ("2.0.0", "sdist", False),
                ("2.5.0", "sdist", False),
                ("1.0.0", "sdist", False),
            ],
            True,
            Version("2.5.0"),
            id="jumbled sdists",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5b0", "sdist", False),
            ],
            True,
            Version("2.0.0"),
            id="pre_lease sdist",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", True),
                ("2.5b0", "sdist", False),
            ],
            True,
            Version("1.0.0"),
            id="pre_release and yanked sdists",
        ),
    ],
)
def test_get_latest_version_sdists(packages_data, non_pre_release, expected):
    packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in packages_data
    ]
    assert get_latest_version(packages, stable=non_pre_release) == expected


@pytest.mark.parametrize(
    "packages_data, non_pre_release, expected",
    [
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "wheel", False),
            ],
            True,
            Version("2.0.0"),
            id="single wheel",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.0.0", "wheel", False),
            ],
            True,
            Version("2.0.0"),
            id="sdist and wheel",
        ),
        pytest.param(
            [
                ("2.0.0", "sdist", False),
                ("2.5.0", "wheel", False),
                ("1.0.0", "sdist", False),
            ],
            True,
            Version("2.0.0"),
            id="jumbled single wheel",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", False),
                ("2.5.0", "wheel", True),
            ],
            True,
            Version("2.0.0"),
            id="single yanked wheel",
        ),
        pytest.param(
            [
                ("1.0.0", "sdist", False),
                ("2.0.0", "sdist", True),
                ("2.5.0", "wheel", False),
            ],
            True,
            Version("1.0.0"),
            id="single wheel yanked sdist",
        ),
    ],
)
def test_get_latest_version_wheels(packages_data, non_pre_release, expected):
    packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in packages_data
    ]
    assert get_latest_version(packages, stable=non_pre_release) == expected


@patch("piplexed.pypi_info.PyPISimple.get_project_page")
def test_get_pypi_versions(mock_page):
    packages_data = [
        ("1.0.0", "sdist", False),
        ("2.0.0", "sdist", False),
        ("2.5.0", "sdist", False),
    ]
    packages = [
        create_autospec(DistributionPackage, version=v, package_type=t, is_yanked=y) for v, t, y in packages_data
    ]

    page = create_autospec(ProjectPage, project="testproj", packages=packages)
    mock_page.return_value = page

    result = get_pypi_versions(session=None, package_name="testproj", stable=True)
    assert list(result) == [PackageInfo(name="testproj", version=Version("2.5.0"), python=None)]


@patch("piplexed.pypi_info.get_pipx_metadata")
@patch("piplexed.pypi_info.get_pypi_versions")
def test_find_outdated_packages(mock_pypi, mock_pipx_metadata, tmp_path):
    mock_pipx_metadata.return_value = [
        PackageInfo(name=canonicalize_name("package_1"), version=Version("1.0.0")),
    ]

    mock_pypi.return_value = [
        PackageInfo(name=canonicalize_name("package_1"), version=Version("2.0.0")),
        PackageInfo(name=canonicalize_name("package_1"), version=Version("1.0.0")),
    ]

    assert find_outdated_packages(tmp_path, stable=True) == [
        PackageVersions(package=canonicalize_name("package_1"), pipx=Version("1.0.0"), pypi=Version("2.0.0")),
    ]


@patch("piplexed.pypi_info.get_pipx_metadata")
@patch("piplexed.pypi_info.get_pypi_versions")
def test_find_outdated_packages_pre(mock_pypi, mock_pipx_metadata, tmp_path):
    mock_pipx_metadata.return_value = [
        PackageInfo(name=canonicalize_name("package_1"), version=Version("1.0.0")),
    ]

    mock_pypi.return_value = [
        PackageInfo(name=canonicalize_name("package_1"), version=Version("2.1b0")),
        PackageInfo(name=canonicalize_name("package_1"), version=Version("2.0.0")),
        PackageInfo(name=canonicalize_name("package_1"), version=Version("1.0.0")),
    ]

    assert find_outdated_packages(tmp_path, stable=False) == [
        PackageVersions(package=canonicalize_name("package_1"), pipx=Version("1.0.0"), pypi=Version("2.1b0")),
    ]



from packaging.version import Version

from piplexed.pipx_venvs import PackageInfo
from piplexed.pipx_venvs import get_pipx_metadata

TEST_PIPX_PYPI_METADATA = """{
    "injected_packages": {},
    "main_package": {
        "app_paths": [
            {
                "__Path__": "C:\\\\path\\\\to\\\\test.exe",
                "__type__": "Path"
            }
        ],
        "app_paths_of_dependencies": {},
        "apps": [
            "test.exe"
        ],
        "apps_of_dependencies": [],
        "include_apps": true,
        "include_dependencies": false,
        "package": "Testy_McTestFace",
        "package_or_url": "Testy_McTestFace",
        "package_version": "23.1.0",
        "pip_args": [],
        "suffix": ""
    },
    "pipx_metadata_version": "0.2",
    "python_version": "Python 3.11.2",
    "venv_args": []
}"""

TEST_PIPX_LOCAL_METADATA = """{
    "injected_packages": {},
    "main_package": {
        "app_paths": [
            {
                "__Path__": "C:\\\\path\\\\to\\\\test.exe",
                "__type__": "Path"
            }
        ],
        "app_paths_of_dependencies": {},
        "apps": [
            "local.exe"
        ],
        "apps_of_dependencies": [],
        "include_apps": true,
        "include_dependencies": false,
        "package": "Local_Package",
        "package_or_url": "path/to/local/wheel",
        "package_version": "23.1.0",
        "pip_args": [],
        "suffix": ""
    },
    "pipx_metadata_version": "0.3",
    "python_version": "Python 3.11.2",
    "venv_args": []
}"""


def test_get_pipx_metadata(tmp_path):
    env_dir = tmp_path / "venvs"
    env_dir.mkdir()
    pypi_package = env_dir / "pypi_package"
    pypi_package.mkdir()
    pypi_test_json = pypi_package / "test.json"
    pypi_test_json.write_text(TEST_PIPX_PYPI_METADATA)

    local_package = env_dir / "local_package"
    local_package.mkdir()
    local_test_json = local_package / "test.json"
    local_test_json.write_text(TEST_PIPX_LOCAL_METADATA)

    assert get_pipx_metadata(env_dir) == [
        PackageInfo(name="testy-mctestface", version=Version("23.1.0"), python="3.11.2"),
    ]



from _pytest.capture import CaptureFixture
from packaging.version import Version

from piplexed._print_tree import print_list_outdated
from piplexed._print_tree import print_list_tree
from piplexed.pipx_venvs import PackageInfo


def test_print_list(capsys: CaptureFixture[str]):
    packages = [
        PackageInfo("A", Version("1.0.1"), python="3.11.3"),
        PackageInfo("B", Version("2.2.1"), python="3.11.3"),
    ]
    print_list_tree(packages)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    assert output[0].startswith("Pip❎ Packages")
    assert output[1] == "├── A"
    assert output[2] == "│   ├── version - 1.0.1"
    assert output[3] == "│   └── python - 3.11.3"


def test_print_list_outdated(capsys: CaptureFixture[str]):
    packages = [
        {"package": "B", "pipx": Version("1.5.0"), "pypi": Version("1.6.0")},
        {"package": "C", "pipx": Version("2.6.0"), "pypi": Version("3.0a0")},
    ]

    print_list_outdated(packages)

    out, err = capsys.readouterr()
    assert not err
    output = [s.strip() for s in out.splitlines()]
    print(out)
    assert output[0] == "Pip❌ Outdated Packages"
    assert output[1] == "├── B"
    assert output[2] == "│   ├── pipx version - 1.5.0"
    assert output[3] == "│   └── PyPI version - 1.6.0"
    assert output[4] == "└── C"
    assert output[5] == "├── pipx version - 2.6.0"
    assert output[6] == "└── PyPI version - 3.0a0 ⚠"



from __future__ import annotations

import typer

import piplexed
from piplexed._print_tree import print_list_outdated
from piplexed._print_tree import print_list_tree

app = typer.Typer()


@app.command()
def version() -> None:
    print(piplexed.__version__)


@app.command()
def list(
    outdated: bool = typer.Option(False, "--outdated", "-O", help="Find outdated packages installed with pipx"),
    is_prelease: bool = typer.Option(
        False,
        "--pre",
        "-P",
        help="Include pre and dev releases in latest pypi version search",
    ),
) -> None:
    if outdated and is_prelease:
        print_list_outdated(piplexed.find_outdated_packages(stable=False))
    elif outdated and not is_prelease:
        print_list_outdated(piplexed.find_outdated_packages(stable=True))
    else:
        print_list_tree(piplexed.get_pipx_metadata())


if __name__ == "__main__":
    raise SystemExit(app)



# SPDX-FileCopyrightText: 2023-present Andrew White <white-aj@outlook.com>
#
# SPDX-License-Identifier: MIT
from .pipx_venvs import get_pipx_metadata
from .pypi_info import find_outdated_packages
from .version import VERSION

__all__ = [
    "get_pipx_metadata",
    "find_outdated_packages",
    "VERSION",
]

__version__ = VERSION



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path

from packaging.utils import NormalizedName
from packaging.utils import canonicalize_name
from packaging.version import Version

DEFAULT_PIPX_HOME = Path.home() / ".local/pipx"
DEFAULT_PIPX_BIN_DIR = Path.home() / ".local/bin"
PIPX_HOME = Path(os.environ.get("PIPX_HOME", DEFAULT_PIPX_HOME)).resolve()
PIPX_LOCAL_VENVS = PIPX_HOME / "venvs"


@dataclass
class PackageInfo:
    name: NormalizedName
    version: Version
    python: str | None = None


def get_pipx_metadata(venv_dir: Path = PIPX_LOCAL_VENVS) -> list[PackageInfo]:
    venvs = []
    for env in venv_dir.iterdir():
        for item in env.iterdir():
            if item.suffix == ".json":  # pragma: no branch
                with open(item) as f:
                    data = json.load(f)
                    # packages installed from pypi have the same package and package_or_url
                    if data["main_package"]["package"] == data["main_package"]["package_or_url"]:
                        pkg_data = PackageInfo(
                            name=canonicalize_name(data["main_package"]["package"]),
                            version=Version(data["main_package"]["package_version"]),
                            python=data["python_version"].split()[-1],
                        )
                        venvs.append(pkg_data)
    return venvs



from __future__ import annotations

from collections.abc import Iterable
from typing import cast

from rich import print as rich_print
from rich.text import Text
from rich.tree import Tree

from piplexed.pipx_venvs import PackageInfo
from piplexed.pypi_info import PackageVersions


def print_list_tree(packages: Iterable[PackageInfo]) -> None:
    tree = Tree("Pip❎ Packages", guide_style="cyan")
    for pkg in packages:
        pkg_name = Text(
            pkg.name.title(),
            "dark_orange bold",
        )
        version_info = Text("version - ", "white").append(f"{pkg.version}", "deep_sky_blue1")
        python_info = Text("python - ", "white").append(cast(str, pkg.python), "dark_green")
        pkg_branch = tree.add(pkg_name)
        pkg_branch.add(version_info)
        pkg_branch.add(python_info)

    rich_print(tree)


def print_list_outdated(package_data: Iterable[PackageVersions]) -> None:
    tree = Tree("Pip❌ Outdated Packages", guide_style="cyan")
    for pkg in package_data:
        pkg_name = Text(
            pkg["package"].title(),
            "bright_yellow bold",
        )

        pipx_info = Text("pipx version - ", "white").append(f"{pkg['pipx']}", "red3")
        pypi_info = Text("PyPI version - ", "white").append(f"{pkg['pypi']}", "green1")
        if pkg["pypi"].is_prerelease or pkg["pypi"].is_devrelease:
            pypi_info.append(" ⚠", "bright_yellow")

        pkg_branch = tree.add(pkg_name)
        pkg_branch.add(pipx_info)
        pkg_branch.add(pypi_info)

    rich_print(tree)



from __future__ import annotations

from collections.abc import Generator
from pathlib import Path
from typing import TypedDict
from typing import cast

from packaging.utils import NormalizedName
from packaging.utils import canonicalize_name
from packaging.version import Version
from platformdirs import user_cache_path
from pypi_simple import DistributionPackage
from pypi_simple import PyPISimple
from requests_cache import CachedSession

from piplexed.pipx_venvs import PackageInfo
from piplexed.pipx_venvs import get_pipx_metadata
from piplexed.version import VERSION

DEFAULT_CACHE: Path = user_cache_path(appname="piplexed", version=VERSION) / "pypi_cache.sqlite"


class PackageVersions(TypedDict):
    package: NormalizedName
    pipx: Version
    pypi: Version


def get_pypi_versions(session: CachedSession, package_name: str, *, stable: bool) -> Generator[PackageInfo, None, None]:
    with PyPISimple(session=session) as client:
        package_page = client.get_project_page(package_name)
        canonicalized_pkg_name = canonicalize_name(package_page.project)

        # use max(Version) instead of reversing order of package_page as recommended in PEP 700
        # https://peps.python.org/pep-0700/

        latest_version = get_latest_version(package_page.packages, stable=stable)

        yield PackageInfo(name=canonicalized_pkg_name, version=latest_version)


def get_latest_version(packages: list[DistributionPackage], *, stable: bool) -> Version:
    if stable:
        latest_version = max(
            pkg_vsn
            for pkg in packages
            if pkg.package_type == "sdist"
            and not (pkg_vsn := Version(cast(str, pkg.version))).is_devrelease
            and not pkg.is_yanked
            and not pkg_vsn.is_prerelease
        )
    else:
        latest_version = max(
            Version(cast(str, pkg.version)) for pkg in packages if pkg.package_type == "sdist" and not pkg.is_yanked
        )

    return latest_version


def find_outdated_packages(cache_dir: Path = DEFAULT_CACHE, *, stable: bool = True) -> list[PackageVersions]:
    updates: list[PackageVersions] = []
    venvs = get_pipx_metadata()
    session = CachedSession(str(cache_dir), backend="sqlite", expire_after=360)
    for pkg in venvs:
        for pypi_release in get_pypi_versions(session, pkg.name, stable=stable):
            if pypi_release.version > pkg.version:
                updates.append({"package": pkg.name, "pipx": pkg.version, "pypi": pypi_release.version})
                break

    session.cache.delete(expired=True)
    return updates



import nox

PYTHON_VERSIONS = ["3.11", "3.10", "3.9", "3.8"]
PYTHON_DEFAULT_VERSION = "3.11"
nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = (
    "tests",
    "coverage_report",
    "lint",
)

DOC_DEPENDENCIES = "requirements/docs.txt"
TEST_DEPENDENCIES = "requirements/tests.txt"
LINT_DEPENDENCIES = "requirements/linting.txt"


@nox.session(python=PYTHON_VERSIONS)
def tests(session):
    session.run("python", "-m", "pip", "install", "-U", "pip")
    session.install(".", "-r", TEST_DEPENDENCIES)
    session.run("coverage", "run", "-m", "pytest", "tests")


@nox.session()
def coverage_report(session):
    session.install("coverage[toml]")

    session.run("coverage", "report", "--show-missing", "--fail-under=80")
    session.run("coverage", "erase")


@nox.session(python=PYTHON_DEFAULT_VERSION)
def lint(session):
    session.run("python", "-m", "pip", "install", "-U", "pip")
    session.install(".", "-r", LINT_DEPENDENCIES)
    session.run("black", "src", "--check")
    session.run("ruff", ".")
    session.run("mypy", "src")


@nox.session(python=PYTHON_DEFAULT_VERSION)
def docs(session):
    session.run("python", "-m", "pip", "install", "-U", "pip")
    session.install(".", "-r", DOC_DEPENDENCIES)
    session.run("mkdocs", "build", "--clean", "--strict")



from packaging.version import parse as parse_version

import piplexed


def test_version():
    v = parse_version(piplexed.VERSION)
    assert str(v) == piplexed.VERSION


def test_version_attribute_is_present():
    assert hasattr(piplexed, "__version__")


def test_version_attribute_is_a_string():
    assert isinstance(piplexed.__version__, str)
